/* Copyright (C) 2022 Art1pirat,

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>

#define FPS 15
#define MAXW 40
#define MAXH 30

/* minitux 
 *
 *  (o>
 *  //\    <-- standard
 *  U_/,
 *
 *  (v)
 *  /U\    <-- front
 *  u-u
 *
 *  (o>
 *  //\    <-- forward
 *  V_,
 *
 *  (o>
 *  //\    <-- no more fish!
 *  U__),
 *
 *  (0>
 *  //\    <-- whoops!
 *  U_/, 
 * 
 *  (o>
 *  //\    <-- travelling
 * [_]/,
 *
 *  Jump animation:
 *
 *  (o>    (o>       (o>     ___
 *  //\  --/\     ---)    ><.___`(o-
 *  V_,   '_-,    <_,      "
 *
 *
 */

unsigned int milliseconds() {
  /*
   * clock_t clocks = clock();
   * return 1000* clocks / CLOCKS_PER_SEC;
   */
  struct timeval time;
  gettimeofday(&time, NULL);
  return  time.tv_sec * 1000 + time.tv_usec / 1000;
}

void frame_sleep() {
  usleep(1000000/FPS);
}

typedef struct sprite_s{
  short int w;
  short int h;
  const char *sprite_str;
} sprite_t;

sprite_t tux_standard () {
  const	sprite_t tux = {
    .w = 4,
    .h = 3,
    .sprite_str =
       "(o> "
       "//\\ "
       "U_/,"
  };
  return tux;
}

sprite_t tux_standard2 () {
  const	sprite_t tux = {
    .w = 4,
    .h = 3,
    .sprite_str =
            "(o> "
            "//\\ "
            "v_,)"

  };
  return tux;
}

sprite_t tux_standard3 () {
  const	sprite_t tux = {
    .w = 4,
    .h = 3,
    .sprite_str =
            "(~> "
            "//\\ "
            "v_,)"
  };
  return tux;
}

sprite_t tux_jump1 () {

  const	sprite_t tux = {
    .w = 5,
    .h = 3,
    .sprite_str =
    "  (o>"
    "--//\\"
    " '_-,"
  };
  return tux;
}

sprite_t tux_jump2 () {
  const	sprite_t tux = {
    .w = 6,
    .h = 3,
    .sprite_str =
            "   (o>"
            "---)  "
            "<_,   "
  };
  return tux;
}

sprite_t tux_jump3 () {
  const	sprite_t tux = {
    .w = 9,
    .h = 3,
    .sprite_str =
            "   ---   "
            "><.__`(o-"
            "\"        "
  };
  return tux;
}

sprite_t cloud1 () {
  const sprite_t cloud = {
    .w = 5,
    .h = 2,
    .sprite_str =
            " .') "
            "(_..)"
  };
  return cloud;
}

sprite_t cloud2 () {
  const sprite_t cloud = {
    .w = 9,
    .h = 4,
    .sprite_str =
    "    _    "
    "  ('#).  "
    "(#####') "
    "  (__#:) "
  };
  return cloud;
}

sprite_t stone1 () {
  const sprite_t stone = {
    .w = 6,
    .h = 2,
    .sprite_str =
        "  _   "
        "//##\\_"
  };
  return stone;
}

sprite_t stone2 () {
  const sprite_t stone = {
    .w = 6,
    .h = 4,
    .sprite_str =
          " ___ "
          "/|  |\\"
          "'.\\/.'"
          "  v   "
  };
  return stone;
}

sprite_t penguin1() {
  const sprite_t penguin1 = {
    .w = 4,
    .h = 2,
    .sprite_str =
    "  _o)"
    ".(\\)"
  };
  return penguin1;

}

sprite_t tree1() {
  const sprite_t tree1 = {
    .w = 4,
    .h = 4,
    .sprite_str =
          " /\\"
          "/  \\"
          "/   \\"
          " `||' "
  };
  return tree1;
}

sprite_t skull() {
  const sprite_t skull = {
    .w = 4,
    .h = 4,
    .sprite_str =
          "  _ "
          " . . "
          "(@-@)"
          " |=| "
  };
  return skull;
}


static const char * sun1[] = {
  "            .           ",
  "                        ",
  "   .        |         . ",
  "                        ",
  "    \\       I       /   ",
  "                        ",
  "      \\  ,g888g  /      ",
  "         d88888b        ",
  "-   --== 8888888  ==-- -",
  "         Y88888P        ",
  "                        ",
  "      /     I     \\     ",
  "                         ",
  "    /       |       \\   ",
  "                         ",
  "  .         .          - "
};

void set_sprite(int x, int y, sprite_t sprite) {
  for (int row = 0; row< sprite.h; row++) {
    for (int col =0; col<sprite.w; col++) {
      char w = sprite.sprite_str[row*sprite.w + col];
      if (w != ' ' && w != '\0') {
        if (w == '#') { w = ' ';} /* for space as foreground */
        mvaddch(row + y, col + x, w);
      }
    }
    //	mvaddstr(i+y, 0+x, s);
  }
}

void display_title_screen() {
  static unsigned int last = 1;
  const	sprite_t title = {
    .w = 37,
    .h = 6,
    .sprite_str =
          "    (o>                              "
          "__  //\\   __ <o))>< _     _ _      _ "
          "\\#\\#U_/,#/ /__#_ #_|#| __|#| | ___|#|"
          " \\#\\#,.#/ / _` |/ _` |/#_`#|#|/#_#\\#|"
          "  \\#V #V#/ (_|#|#(_|#|#(_|#|#|##__/_|"
          "   \\_/\\_/#\\__,_|\\__,_|\\__,_|_|\\___(_)"
  };
  set_sprite(0,0, title);
  if ((last++ % (1700*FPS/1000)) == 0) {
    mvaddch(0,5, '~');
  } else {
    mvaddch(0,5, 'o');
  }
  mvaddstr(10, 0, "Press key:");
  mvaddstr(11, 0, "  <q>      to exit,");
  mvaddstr(12, 0, "<SPACE>    to start/jump.");
  mvaddstr(15, 0, "(c) 2022 by Art1Pirat, licensed under GPL 3.0.");
}

unsigned int wait_frames(unsigned int speed) {
  /* 25fps, speed1: 25fps
   *        speed100: 5fps
   */
  // 1000/25 ms = 4ms => 1 Frame / 25fps
  // 1000/5 ms 200ms  => 1 Frame / 5 fps

  return 100 / FPS * speed;
}

int animate_tux (int x, int y, int jump, unsigned int speed) {
  static short unsigned int last = 0;
  static unsigned int last_wait_point=0;
  static unsigned short int in_jump = 0;
  if (in_jump == 0) {
    if (jump) { in_jump=7; }
    if ((last % 2) == 1) {
      set_sprite(x, y, tux_standard());
    } else {
      set_sprite(x, y, tux_standard2());
    }
    if ((last % 7) == 0) {
      set_sprite(x, y, tux_standard3());
    }
  } else {
    switch (in_jump) {
      case 7 :
      case 1 :
        set_sprite(x, y-1, tux_jump1());
        break;
      case 6:
      case 2:
        set_sprite(x,y-3, tux_jump2());
        break;
      case 5:
      case 4:
      case 3:
        set_sprite(x,y-6, tux_jump3());
        break;
    }
  }
  if ( last_wait_point < milliseconds() - wait_frames(speed)  ) {
    last++;
    last_wait_point = milliseconds();
    if (in_jump>0) in_jump--;
  }
  return (in_jump>0);
}

int min(int a, int b) {
  return (a<b?a:b);
}

void animate_ground_line(int x, int y, unsigned int speed) {
  const char ground[] = "~~-~-----~-------~------~-~~-~-----~-------~------~-~~";
  int maxshift = min(MAXW, strlen(ground)-1);
  static int last = 0;
  static unsigned last_wait_point = 0;
  int pos = 0;
  /* render from last to maxshift */
  for(int i = last; i <= maxshift; i++) {
    mvaddch(0+y, pos+x, ground[i]);
    pos++;
  }
  /* render from 0 to last -1 */
  for(int i = 0; i < last; i++) {
    mvaddch(0+y, pos+x, ground[i]);
    pos++;
  }
  /* 25fps, speed1: 25fps
   *        speed100: 5fps
   */
  /* DEBUG OUTPUT: 
     char tmp[200];
     sprintf(tmp, "lastw=%u w=%u, p=%u s=%u", last_wait_point, wait_frames(speed), points, speed);
     mvaddstr(30, 30, tmp);
   */ 
  if ( last_wait_point < milliseconds() - wait_frames(speed)  ) {
    last++;
    last_wait_point = milliseconds();
  }
  if (last > maxshift) { last = 0; }
}

unsigned int animate_targets(int x, int y, unsigned int speed) {
  static int last = MAXW;
  static unsigned int last_wait_point = 0;
  unsigned newx = x+last;
  set_sprite(newx, y, stone1());
  if ( last_wait_point < milliseconds() - wait_frames(speed)  ) {
    last--;
    last_wait_point = milliseconds();
  }
  if (last + stone1().w < 1) {
    last = MAXW;
  }
  return newx;
}

void animate_sun(int x, int y) {
  for(int i = 0; i < 16; i++) {
    mvaddstr(y+i, 0+x, sun1[i]);
  }
}


void animate_clouds(int x, int y) {
  static int last_cloud1_h = 0;
  static int last_cloud2_h = 0;
  static int last_cloud1_x = 20;
  /* render cloud 1 */
  set_sprite(x+last_cloud1_x,y+last_cloud1_h, cloud1());
  static int last_cloud2_x = 26;
  /* render cloud 2 */
  set_sprite(x+last_cloud2_x,y+last_cloud2_h, cloud2());
  last_cloud1_x--;
  last_cloud2_x--;
  if ((last_cloud1_x+cloud1().w) < 1) {
    last_cloud1_x=MAXW;
    last_cloud1_h = rand() % 4;
  }
  if ((last_cloud2_x+cloud2().w) < 1) {
    last_cloud2_x=MAXW;
    last_cloud2_h = rand() % 5;
  }
}

void draw_stats(int x, int y, unsigned int points) {
  char tmp[200];
  sprintf(tmp, "%u points", points/10);
  mvaddstr(y, x, tmp);
}

void start_screen() {
  int ch;
  do {
    display_title_screen();
    refresh();
    frame_sleep();
    ch = getch();
  } while ((ch != 'q') && (ch != 'Q') && (ch != ' '));
  erase();
}

void single_game() {
  int ch;
  int pause = 0;
  unsigned int speed = 25;
  unsigned int pos_target = MAXW;
  int gameover = 0;
  int tux_is_jumping = 0;
  unsigned int points = 0;
  do {
    ch = getch();
    /* test inputted key and determine direction */
    if (pause && (ch != 'q') && (ch != 'Q')) { }
    else {
      erase();
      animate_sun(12, 0);
      animate_clouds(0, 0);
      animate_ground_line(5, 2 + 10, speed);
      draw_stats(18, 0, points);
      if (!gameover) {
        pos_target = animate_targets(5, 11, speed);
	if (!pause) points++;
      }
      if (gameover || (!tux_is_jumping && pos_target <= 4)) { /* 4 is width of tux */
        mvaddstr(11,10, " [ G A M E   O V E R ! ] ");
        refresh();
        gameover = 1;
      }
    }
    switch (ch) {
      //case KEY_UP:
      //animate_tux(0,0+8);
      //    break;
      case KEY_DOWN:
        break;
      case KEY_LEFT:
        break;
      case KEY_RIGHT:
        //animate_tux(0,0+10);
        break;
      case 'p':
        pause = abs(pause -1);
        if (pause) {
          mvaddstr(10,10, " [ P A U S E ] ");
          refresh();
	}
        break;
      case ' ':
      case KEY_UP:
        if (!pause) {
          tux_is_jumping = animate_tux(0, 0 + 10, TRUE, speed);
        }
        break;
      default:
        if (!pause) {
          tux_is_jumping = animate_tux(0, 0 + 10, FALSE, speed);
        }
    }
    if (0 == ( points % 100 ) ) { speed--; }
    if (speed < 1) { speed = 1;}
    if (!pause) refresh();
    frame_sleep();
  } while ((ch != 'q') && (ch != 'Q') );
  erase();
}


int main(int argc, char *argv[])
{
  initscr();
  cbreak(); /* disable linebuffering */
  noecho(); /* disable echo of input */
  keypad(stdscr, TRUE); /* enables cursor keys */
  nodelay(stdscr, TRUE); /* disables blocking waiting for keypress */
  idcok(stdscr, FALSE); /* disables character insert/delete */
  leaveok(stdscr, TRUE); /* disables cursor refresh */
  scrollok(stdscr, FALSE); /* disable softscrolling */
  start_screen();
  single_game();
  endwin();
  return 0;
}


/* vim: set tabstop=2 softtabstop=2 shiftwidth=2 smarttab expandtab :*/
